#include "OgreOcclusionCullingTerrainSceneTraverser.h"
#include "OgreOcclusionCullingTerrainSceneManager.h"
#include "OgreHardwareOcclusionQuery.h"

#include <windows.h>

namespace Ogre {
	//-----------------------------------------------------------------------
	OcclusionCullingTerrainSceneTraverser::OcclusionCullingTerrainSceneTraverser(SceneManager *sm, RenderSystem *rsys): 
	OcclusionCullingSceneTraverser(sm, rsys)
	{
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingTerrainSceneTraverser::traverseOctant(Camera *cam, Octree *octant )
	{
		mNumTraversedNodes ++;
	
		//if(octant->numNodes() > 0)
			renderOctant(cam, octant);

		for(int i=0; i<8; i++)
		{
			Octree *nextChild = octant->mChildren[(i & 4) >> 2][(i & 2) >> 1][i & 1];
		
			if(nextChild != 0)
				mDistanceQueue->push(nextChild);
		}
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingTerrainSceneTraverser::renderScene( Camera *cam, Octree *root )
		//TODO: Renderscene independent from hierarchynode => implement only once
	{
		mNumTraversedNodes = 0;
		mNumQueryCulledNodes = 0;
		mNumFrustumCulledNodes = 0;
		mNumRenderedGeometry = 0;
		
		mDistanceQueue = new OctreePriorityQueue(octreeless<Octree *>(cam));
		mDistanceQueue->push(root);
		mCurrentTestIdx = 0;
		
//mCurrentAlgorithm = RENDER_CULL_FRUSTUM;
		switch(mCurrentAlgorithm)
		{
			case RENDER_CULL_FRUSTUM:
				renderCullFrustum(cam);
		        break;
			case RENDER_STOP_AND_WAIT:
				renderStopAndWait(cam);
				break;
			case RENDER_COHERENT:
				//renderCoherentWithQueue(cam);
				renderStopAndWait(cam);
				break;
			default:
				renderCullFrustum(cam);
				break;
		}	
		
		delete mDistanceQueue;

		mFrameId ++;
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingTerrainSceneTraverser::renderCullFrustum(Camera *cam)
	{
		AxisAlignedBox box;

		while(!mDistanceQueue->empty())
		{
			Octree *octant = mDistanceQueue->top();
			mDistanceQueue->pop();
	
			// interesting for visualization purpose
			//TODO: octree->setNodeVisible(false);
			octant->_getCullBounds(&box);
			
			if(static_cast<OctreeCamera *>(cam)->getVisibility(box) == OctreeCamera::NONE)
			if(!cam->isVisible(box))
			{
				mNumFrustumCulledNodes ++;
				continue;
			}

			// update node's visited flag 
			//TODO: octree->setLastVisited(mFrameId);
			//TODO: octree->setNodeVisible(true);
		
			traverseOctant(cam, octant);
		}
	}	
	//-----------------------------------------------------------------------
	/** Renders the scene with the hierarchical stop and wait algorithm. */
	void OcclusionCullingTerrainSceneTraverser::renderStopAndWait( Camera *cam )
	{
		AxisAlignedBox box;

		while(!mDistanceQueue->empty())
		{
			Octree *octant = mDistanceQueue->top();
			mDistanceQueue->pop();
		
			// interesting for visualization purpose
			//octant->setNodeVisible(false);
			//octant->setLastVisited(mFrameId);

			//TODO: Isvisible also checked inside scenenode::findvisibleobjects
			octant->_getCullBounds(&box);

			char msg[100]; 
			Vector3 max = box.getMaximum();
			Vector3 min = box.getMinimum();

			sprintf(msg, "box max: %3.3f, %3.3f, %3.3f, min: %3.3f, %3.3f, %3.3f\n",max.x,max.y,max.z,min.x,min.y,min.z);
			OutputDebugString(msg);

			if(static_cast<OctreeCamera *>(cam)->getVisibility(box) == OctreeCamera::NONE)
			//if(!cam->isVisible(box))
			{
				mNumFrustumCulledNodes ++;
				continue;
			}

			HardwareOcclusionQuery *query = issueOcclusionQuery(&box, false);
				
			unsigned int visiblePixels;
			// wait if result not available
			query->pullOcclusionQuery(&visiblePixels);
//traverseOctant(cam, octant);			
			// node visible
			if(visiblePixels > mVisibilityThreshold)
			{
				traverseOctant(cam, octant);
				//mNumQueryCulledNodes ++;
			}
			else
			{
				mNumQueryCulledNodes ++;
			}
		}
	}
	//-----------------------------------------------------------------------
	/** Renders the scene with the coherent hierarchical algorithm and the query queye. */
	void OcclusionCullingTerrainSceneTraverser::renderCoherentWithQueue( Camera *cam )
	{
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingTerrainSceneTraverser::pullUpVisibility( Octree *node )
	{
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingTerrainSceneTraverser::renderOctant( Camera *cam, Octree *octree )
	{
		setRenderingMode(MODE_RENDER);
		((OcclusionCullingTerrainSceneManager *)mSceneManager)->_renderOctant(cam, octree);
	}
}	